"""Test advertisements for aiobmsble package.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from functools import cache
from importlib import resources
import json
from string import hexdigits
from typing import Any

from bleak.backends.scanner import AdvertisementData

type BmsAdvList = list[tuple[AdvertisementData, str, str, list[str]]]


def adv_dict_to_advdata(adv_dict: dict[str, Any]) -> AdvertisementData:
    """Generate an AdvertisementData instance from a JSON dictionary."""
    ADVERTISEMENT_DATA_DEFAULTS: dict[str, Any] = {
        "local_name": None,
        "manufacturer_data": {},
        "service_data": {},
        "service_uuids": [],
        "rssi": -127,
        "platform_data": ((),),
        "tx_power": None,
    }

    if "manufacturer_data" in adv_dict:
        adv_dict["manufacturer_data"] = {
            int(k): bytes.fromhex(v) if isinstance(v, str) else v
            for k, v in adv_dict["manufacturer_data"].items()
        }
    if "service_data" in adv_dict:
        adv_dict["service_data"] = {
            k: bytes.fromhex(v) if isinstance(v, str) else v
            for k, v in adv_dict["service_data"].items()
        }
    if "platform_data" in adv_dict:
        pdata = adv_dict.get("platform_data", [])
        assert isinstance(
            pdata[0], str
        ), "first entry of platform_data needs to be string."
        parts = pdata[0].split(":")
        assert len(parts) == 6 and all(
            len(part) == 2 and all(c in hexdigits for c in part) for part in parts
        ), "first entry of platform_data only accepts MAC addresses"
        adv_dict["platform_data"] = tuple(pdata)

    return AdvertisementData(**(ADVERTISEMENT_DATA_DEFAULTS | adv_dict))


@cache
def bms_advertisements(bms_filter: str | None = None) -> BmsAdvList:
    """Provide all available BMS advertisements from test data directory.

    Load *_bms.json files from the packaged test data directory.
    If no type is given, all available advertisemetns (any type) will be returned.

    Args:
        bms_filter (str|None): BMS type for which the advertisements should be returned.

    Returns:
        BmsAdvList: List of tuples containing advertisement, mac_addr, bms type,
        and a list of comments, i.e. list[tuple[AdvertisementData, str, list[str]]]

    """
    all_data: BmsAdvList = []

    for resource in resources.files(__package__).iterdir():
        if bms_filter and not resource.name.startswith(bms_filter):
            continue
        if not resource.name.endswith("_bms.json"):
            continue
        with resource.open("r", encoding="UTF-8") as f:
            raw_data: Any = json.load(f)
            assert isinstance(raw_data, list)

            for entry in raw_data:
                assert isinstance(entry, dict)
                assert {"advertisement", "type"}.issubset(set(entry.keys()))
                adv: AdvertisementData = adv_dict_to_advdata(entry["advertisement"])
                mac_addr: str = (
                    adv.platform_data[0]
                    if isinstance(adv.platform_data, tuple)
                    and isinstance(adv.platform_data[0], str)
                    else ""
                )
                bms_type: str = entry["type"]
                comments: list[str] = entry["_comments"]

                assert isinstance(adv, AdvertisementData)
                assert isinstance(bms_type, str)
                assert isinstance(comments, list)
                assert all(isinstance(c, str) for c in comments)
                assert resource.name == f"{bms_type}.json"

                all_data.append((adv, mac_addr, bms_type, comments))
    return all_data


def ignore_advertisements() -> BmsAdvList:
    """Provide a list of advertisements that shall not be identified as a valid BMS.

    Load ignore.json files from the packaged test data directory.

    Returns:
        BmsAdvList: List of tuples containing advertisement, reason why not to detect,
        and a list of comments, i.e. list[tuple[AdvertisementData, str, list[str]]]

    """
    data: BmsAdvList = []

    with (
        resources.files(__package__)
        .joinpath("ignore.json")
        .open("r", encoding="UTF-8") as f
    ):
        raw_data: Any = json.load(f)
        assert isinstance(raw_data, list)

        for entry in raw_data:
            assert isinstance(entry, dict)
            assert {"advertisement", "reason"}.issubset(set(entry.keys()))
            adv: AdvertisementData = adv_dict_to_advdata(entry["advertisement"])
            mac_addr: str = adv.platform_data[0] if "platform" in adv else ""
            reason: str = entry["reason"]
            comments: list[str] = entry["_comments"]

            assert isinstance(adv, AdvertisementData)
            assert isinstance(reason, str)
            assert isinstance(comments, list)
            assert all(isinstance(c, str) for c in comments)

            data.append((adv, mac_addr, reason, comments))

    return data



"""Module to support Seplos V3 smart BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from collections.abc import Callable
from functools import cache
from typing import Any, Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSpackvalue, BMSSample, MatcherPattern
from aiobmsble.basebms import BaseBMS, crc_modbus, swap32


class BMS(BaseBMS):
    """Seplos V3 smart BMS class implementation."""

    INFO: BMSInfo = {"default_manufacturer": "Seplos", "default_model": "smart BMS V3"}
    _CMD_READ: Final[list[int]] = [0x01, 0x04]
    _HEAD_LEN: Final[int] = 3
    _CRC_LEN: Final[int] = 2
    _PIA_LEN: Final[int] = 0x11
    _PIB_LEN: Final[int] = 0x1A
    _EIA_LEN: Final[int] = _PIB_LEN
    _EIB_LEN: Final[int] = 0x16
    _EIC_LEN: Final[int] = 0x5
    _TEMP_START: Final[int] = _HEAD_LEN + 32
    _QUERY: Final[dict[str, tuple[int, int, int]]] = {
        # name: cmd, reg start, length
        "EIA": (0x4, 0x2000, _EIA_LEN),
        "EIB": (0x4, 0x2100, _EIB_LEN),
        "EIC": (0x1, 0x2200, _EIC_LEN),
    }
    _PQUERY: Final[dict[str, tuple[int, int, int]]] = {
        "PIA": (0x4, 0x1000, _PIA_LEN),
        "PIB": (0x4, 0x1100, _PIB_LEN),
    }
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("temperature", 20, 2, True, lambda x: x / 10, _EIB_LEN),  # avg. ctemp
        BMSDp("voltage", 0, 4, False, lambda x: swap32(x) / 100, _EIA_LEN),
        BMSDp("current", 4, 4, True, lambda x: swap32(x, True) / 10, _EIA_LEN),
        BMSDp("cycle_charge", 8, 4, False, lambda x: swap32(x) / 100, _EIA_LEN),
        BMSDp("pack_count", 44, 2, False, idx=_EIA_LEN),
        BMSDp("cycles", 46, 2, False, idx=_EIA_LEN),
        BMSDp("battery_level", 48, 2, False, lambda x: x / 10, _EIA_LEN),
        BMSDp("battery_health", 50, 2, False, lambda x: x / 10, _EIA_LEN),
        BMSDp("problem_code", 1, 9, False, lambda x: x & 0xFFFF00FF00FF0000FF, _EIC_LEN),
        BMSDp("dischrg_mosfet", 7, 1, False, lambda x: bool(x & 1), _EIC_LEN),
        BMSDp("chrg_mosfet", 7, 1, False, lambda x: bool(x & 2), _EIC_LEN),
        BMSDp("heater", 7, 1, False, lambda x: bool(x & 8), _EIC_LEN),
        BMSDp("balancer", 7, 1, False, lambda x: bool(x & 4), _EIC_LEN),  # limit FET
    )  # Protocol Seplos V3
    _PFIELDS: Final[
        tuple[tuple[BMSpackvalue, int, bool, Callable[[int], Any]], ...]
    ] = (
        ("pack_voltages", 0, False, lambda x: x / 100),
        ("pack_currents", 2, True, lambda x: x / 100),
        ("pack_battery_levels", 10, False, lambda x: x / 10),
        ("pack_battery_health", 12, False, lambda x: x / 10),
        ("pack_cycles", 14, False, lambda x: x),
    )  # Protocol Seplos V3
    _CMDS: Final = frozenset(
        {field[2] for field in _QUERY.values()}
        | {field[2] for field in _PQUERY.values()}
    )

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize private BMS members."""
        super().__init__(ble_device, keep_alive)
        self._msg: dict[int, bytes] = {}
        self._pack_count: int = 0  # number of battery packs
        self._pkglen: int = 0  # expected packet length

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {
                "local_name": pattern,
                "service_uuid": BMS.uuid_services()[0],
                "connectable": True,
            }
            for pattern in ("SP??B*", "XZHX*", "CSY*", "SP1??B*")
        ]

    # setup UUIDs
    #    serv 0000fff0-0000-1000-8000-00805f9b34fb
    # 	 char 0000fff1-0000-1000-8000-00805f9b34fb (#16): ['read', 'notify']
    # 	 char 0000fff2-0000-1000-8000-00805f9b34fb (#20): ['read', 'write-without-response', 'write']
    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return (normalize_uuid_str("fff0"),)

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "fff1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "fff2"

    # async def _fetch_device_info(self) -> BMSInfo: use default, VIA msg useless

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Retrieve BMS data update."""

        if (
            len(data) > BMS._HEAD_LEN + BMS._CRC_LEN
            and data[0] <= self._pack_count
            and data[1] & 0x7F in BMS._CMD_READ  # include read errors
            and data[2] >= BMS._HEAD_LEN + BMS._CRC_LEN
        ):
            self._frame.clear()
            self._pkglen = data[2] + BMS._HEAD_LEN + BMS._CRC_LEN
        elif (  # error message
            len(data) == BMS._HEAD_LEN + BMS._CRC_LEN
            and data[0] <= self._pack_count
            and data[1] & 0x80
        ):
            self._log.debug("RX error: %X", data[2])
            self._frame.clear()
            self._pkglen = BMS._HEAD_LEN + BMS._CRC_LEN

        self._frame += data
        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._frame else "cnt.", data
        )

        # verify that data is long enough
        if len(self._frame) < self._pkglen:
            return

        if (crc := crc_modbus(self._frame[: self._pkglen - 2])) != int.from_bytes(
            self._frame[self._pkglen - 2 : self._pkglen], "little"
        ):
            self._log.debug(
                "invalid checksum 0x%X != 0x%X",
                int.from_bytes(self._frame[self._pkglen - 2 : self._pkglen], "little"),
                crc,
            )
            self._frame.clear()
            return

        if self._frame[2] >> 1 not in BMS._CMDS or self._frame[1] & 0x80:
            self._log.debug(
                "unknown message: %s, length: %s", self._frame[0:2], self._frame[2]
            )
            self._frame.clear()
            return

        if len(self._frame) != self._pkglen:
            self._log.debug(
                "wrong data length (%i!=%s): %s",
                len(self._frame),
                self._pkglen,
                self._frame,
            )

        self._msg[self._frame[0] << 8 | self._frame[2] >> 1] = bytes(self._frame)
        self._frame.clear()
        self._msg_event.set()

    async def _init_connection(
        self, char_notify: BleakGATTCharacteristic | int | str | None = None
    ) -> None:
        """Initialize RX/TX characteristics."""
        await super()._init_connection()
        self._pack_count = 0
        self._pkglen = 0

    @staticmethod
    @cache
    def _cmd(device: int, cmd: int, start: int, count: int) -> bytes:
        """Assemble a Seplos BMS command."""
        assert device >= 0x00 and (device <= 0x10 or device in (0xC0, 0xE0))
        assert cmd in (0x01, 0x04)  # allow only read commands
        assert start >= 0 and count > 0 and start + count <= 0xFFFF
        frame: bytearray = bytearray([device, cmd])
        frame += int.to_bytes(start, 2, byteorder="big")
        frame += int.to_bytes(count * (0x10 if cmd == 0x1 else 0x1), 2, byteorder="big")
        frame += int.to_bytes(crc_modbus(frame), 2, byteorder="little")
        return bytes(frame)

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        for block in BMS._QUERY.values():
            await self._await_msg(BMS._cmd(0x0, *block))

        data: BMSSample = BMS._decode_data(BMS._FIELDS, self._msg, start=BMS._HEAD_LEN)

        self._pack_count = min(data.get("pack_count", 0), 0x10)

        for pack in range(1, 1 + self._pack_count):
            for block in BMS._PQUERY.values():
                await self._await_msg(self._cmd(pack, *block))

            for key, idx, sign, func in BMS._PFIELDS:
                data.setdefault(key, []).append(
                    func(
                        int.from_bytes(
                            self._msg[pack << 8 | BMS._PIA_LEN][
                                BMS._HEAD_LEN + idx : BMS._HEAD_LEN + idx + 2
                            ],
                            byteorder="big",
                            signed=sign,
                        )
                    )
                )

            pack_cells: list[float] = BMS._cell_voltages(
                self._msg[pack << 8 | BMS._PIB_LEN], cells=16, start=BMS._HEAD_LEN
            )
            # update per pack delta voltage
            data["delta_voltage"] = max(
                data.get("delta_voltage", 0),
                round(max(pack_cells) - min(pack_cells), 3),
            )
            # add individual cell voltages
            data.setdefault("cell_voltages", []).extend(pack_cells)
            # add temperature sensors (4x cell temperature + 4 reserved)
            data.setdefault("temp_values", []).extend(
                BMS._temp_values(
                    self._msg[pack << 8 | BMS._PIB_LEN],
                    values=4,
                    start=BMS._TEMP_START,
                    signed=False,
                    offset=2731,
                    divider=10,
                )
            )
            # calculate cell_count instead of querying SPA
            data["cell_count"] = len(data.get("cell_voltages", [])) // self._pack_count

        self._msg.clear()

        return data



"""Module to support LiPower BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from functools import cache
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, MatcherPattern
from aiobmsble.basebms import BaseBMS, crc_modbus


class BMS(BaseBMS):
    """LiPower BMS implementation."""

    INFO: BMSInfo = {"default_manufacturer": "Ective", "default_model": "LiPower BMS"}
    _HEADS: Final[tuple[bytes, ...]] = (b"\x22\x03", b"\x0b\x03")  # alternative heads
    _MIN_LEN: Final[int] = 5  # minimal frame length, including SOF and checksum
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("voltage", 15, 2, False, lambda x: x / 10),
        BMSDp(
            "current", 12, 3, False, lambda x: (x & 0xFFFF) * -(1 ** (x >> 16)) / 100
        ),
        BMSDp("battery_level", 5, 2, False),
        BMSDp(
            "runtime",
            7,
            4,
            False,
            lambda x: (x >> 16) * BMS._HRS_TO_SECS + (x & 0xFFFF) * 60,
        ),
        BMSDp("cycle_charge", 3, 2, False),
        # BMSDp("power", 17, 2, False),  # disabled, due to precision
    )

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._heads: tuple[bytes, ...] = BMS._HEADS
        self._msg: bytes = b""

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [{"service_uuid": normalize_uuid_str("af30"), "connectable": True}]

    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return (normalize_uuid_str("ffe0"),)

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "ffe1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "ffe1"

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug("RX BLE data: %s", data)

        if not data.startswith(self._heads) or len(data) < BMS._MIN_LEN:
            self._log.debug("incorrect SOF")
            return

        if len(data) != data[2] + BMS._MIN_LEN:
            self._log.debug("incorrect frame length")
            return

        if (crc := crc_modbus(data[:-2])) != int.from_bytes(
            data[-2:], byteorder="little"
        ):
            self._log.debug(
                "invalid checksum 0x%X != 0x%X",
                int.from_bytes(data[-2:], byteorder="little"),
                crc,
            )
            return

        self._msg = bytes(data)
        self._msg_event.set()

    @staticmethod
    @cache
    def _cmd(cmd: int, addr: int, words: int, head: bytes) -> bytes:
        """Assemble a LiPower BMS command (MODBUS)."""
        frame: bytearray = (
            bytearray(head)
            + cmd.to_bytes(1, "big")
            + addr.to_bytes(2, "big")
            + words.to_bytes(1, "big")
        )
        frame.extend(int.to_bytes(crc_modbus(frame), 2, byteorder="little"))
        return bytes(frame)

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        for head in self._heads:
            try:
                await self._await_msg(
                    BMS._cmd(cmd=0x4, addr=0x0, words=0x8, head=head)
                )
                if len(self._heads) > 1:
                    self._log.debug("detected frame head: %s", head.hex(" "))
                    self._heads = (head,)  # set to single head for further commands
                break
            except TimeoutError:
                ...  # try next frame head
        else:
            raise TimeoutError

        return BMS._decode_data(BMS._FIELDS, self._msg)



"""Module to support Seplos v2 BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from functools import cache
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, MatcherPattern
from aiobmsble.basebms import BaseBMS, b2str, crc_xmodem


class BMS(BaseBMS):
    """Seplos v2 BMS implementation."""

    INFO: BMSInfo = {"default_manufacturer": "Seplos", "default_model": "smart BMS V2"}
    _HEAD: Final[bytes] = b"\x7e"
    _TAIL: Final[bytes] = b"\x0d"
    _CMD_VER: Final[int] = 0x10  # TX protocol version
    _RSP_VER: Final[int] = 0x14  # RX protocol version
    _MIN_LEN: Final[int] = 10
    _MAX_SUBS: Final[int] = 0xF
    _CELL_POS: Final[int] = 9
    _PRB_MAX: Final[int] = 8  # max number of alarm event bytes
    _PRB_MASK: Final[int] = 0x7DFFFFFFFFFF  # ignore byte 7-8 + byte 6 (bit 7,2)
    _PFIELDS: Final[tuple[BMSDp, ...]] = (  # Seplos V2: single machine data
        BMSDp("voltage", 2, 2, False, lambda x: x / 100),
        BMSDp("current", 0, 2, True, lambda x: x / 100),  # /10 for 0x62
        BMSDp("cycle_charge", 4, 2, False, lambda x: x / 100),  # /10 for 0x62
        BMSDp("cycles", 13, 2, False),
        BMSDp("battery_level", 9, 2, False, lambda x: x / 10),
        BMSDp("battery_health", 15, 2, False, lambda x: x / 10),
    )
    _GSMD_LEN: Final[int] = _CELL_POS + max((dp.pos + dp.size) for dp in _PFIELDS) + 3
    _CMDS: Final[frozenset[tuple[int, bytes]]] = frozenset(
        {(0x51, b""), (0x61, b"\x00"), (0x62, b"")}
    )

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._msg: dict[int, bytes] = {}
        self._exp_len: int = BMS._MIN_LEN
        self._exp_reply: set[int] = set()

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {
                "local_name": pattern,
                "service_uuid": BMS.uuid_services()[0],
                "connectable": True,
            }
            for pattern in ("BP0?", "BP1?", "BP2?")
        ]

    @staticmethod
    def uuid_services() -> tuple[str, ...]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return (normalize_uuid_str("ff00"),)

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "ff01"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "ff02"

    async def _fetch_device_info(self) -> BMSInfo:
        """Fetch the device information via BLE."""
        self._exp_reply = {0x51}
        await self._await_msg(BMS._cmd(0x51))
        _dat: Final[bytes] = self._msg[0x51]
        return {
            "model": b2str(_dat[26:36]),
            "sw_version": f"{int(_dat[37])}.{int(_dat[38])}",
        }

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        if (
            len(data) > BMS._MIN_LEN
            and data.startswith(BMS._HEAD)
            and len(self._frame) >= self._exp_len
        ):
            self._exp_len = BMS._MIN_LEN + int.from_bytes(data[5:7])
            self._frame = bytearray()

        self._frame += data
        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._frame else "cnt.", data
        )

        # verify that data is long enough
        if len(self._frame) < self._exp_len:
            return

        if not self._frame.endswith(BMS._TAIL):
            self._log.debug("incorrect frame end: %s", self._frame)
            return

        if self._frame[1] != BMS._RSP_VER:
            self._log.debug("unknown frame version: V%.1f", self._frame[1] / 10)
            return

        if self._frame[4]:
            self._log.debug("BMS reported error code: 0x%X", self._frame[4])
            return

        if (crc := crc_xmodem(self._frame[1:-3])) != int.from_bytes(self._frame[-3:-1]):
            self._log.debug(
                "invalid checksum 0x%X != 0x%X",
                crc,
                int.from_bytes(self._frame[-3:-1]),
            )
            return

        self._log.debug(
            "address: 0x%X, function: 0x%X, return: 0x%X",
            self._frame[2],
            self._frame[3],
            self._frame[4],
        )

        self._msg[self._frame[3]] = bytes(self._frame)
        try:
            self._exp_reply.remove(self._frame[3])
            self._msg_event.set()
        except KeyError:
            self._log.debug("unexpected reply: 0x%X", self._frame[3])

    async def _init_connection(
        self, char_notify: BleakGATTCharacteristic | int | str | None = None
    ) -> None:
        """Initialize protocol state."""
        await super()._init_connection()
        self._exp_len = BMS._MIN_LEN

    @staticmethod
    @cache
    def _cmd(cmd: int, address: int = 0, data: bytes = b"") -> bytes:
        """Assemble a Seplos V2 BMS command."""
        assert cmd in (0x47, 0x51, 0x61, 0x62, 0x04)  # allow only read commands
        frame = bytearray([*BMS._HEAD, BMS._CMD_VER, address, 0x46, cmd])
        frame += len(data).to_bytes(2, "big", signed=False) + data
        frame += int.to_bytes(crc_xmodem(frame[1:]), 2, byteorder="big") + BMS._TAIL
        return bytes(frame)

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""

        for cmd, data in BMS._CMDS:
            self._exp_reply.add(cmd)
            await self._await_msg(BMS._cmd(cmd, data=data))

        result: BMSSample = {}
        result["cell_count"] = self._msg[0x61][BMS._CELL_POS]
        result["temp_sensors"] = self._msg[0x61][
            BMS._CELL_POS + result["cell_count"] * 2 + 1
        ]
        ct_blk_len: Final[int] = (result["cell_count"] + result["temp_sensors"]) * 2 + 2

        if (BMS._GSMD_LEN + ct_blk_len) > len(self._msg[0x61]):
            raise ValueError("message too short to decode data")

        result |= BMS._decode_data(
            BMS._PFIELDS, self._msg[0x61], start=BMS._CELL_POS + ct_blk_len
        )

        # get extension pack count from parallel data (main pack)
        result["pack_count"] = self._msg[0x51][42]

        # get switches from parallel data (main pack)
        states: Final[int] = self._msg[0x62][45]
        result |= {
            "dischrg_mosfet": bool(states & 0x1),
            "chrg_mosfet": bool(states & 0x2),
            "balancer": bool(states & 0x4),
            "heater": bool(states & 0x8),
        }

        # get alarms from parallel data (main pack)
        alarm_evt: Final[int] = min(self._msg[0x62][46], BMS._PRB_MAX)
        result["problem_code"] = (
            int.from_bytes(self._msg[0x62][47 : 47 + alarm_evt], byteorder="big")
            & BMS._PRB_MASK
        )

        result["cell_voltages"] = BMS._cell_voltages(
            self._msg[0x61],
            cells=self._msg[0x61][BMS._CELL_POS],
            start=10,
        )
        result["temp_values"] = BMS._temp_values(
            self._msg[0x61],
            values=result.get("temp_sensors", 0),
            start=BMS._CELL_POS + result.get("cell_count", 0) * 2 + 2,
            signed=False,
            offset=2731,
            divider=10,
        )

        self._msg.clear()

        return result
